# -*- coding: utf-8 -*-

"""Top-level package for Python Brain Signal Analysis."""

__author__ = """Abdulrahman A. Althobaiti"""
__email__ = 'Abdulrahman.Althobaiti@kaust.edu.sa'
__version__ = '0.1.0'

#from ._version import get_versions
#__version__ = get_versions()['version']
#del get_versions
from .loader import loader as loader